Package: polle
Title: Policy Learning
Version: 1.6.4
Authors@R: c(person(given = "Andreas",
                    family = "Nordland",
                    role = c("aut", "cre"),
                    email = "andreasnordland@gmail.com"),
             person(given = "Klaus",
                    family = "Holst",
                    role = c("aut"),
                    email = "klaus@holst.it",
                    comment = c(ORCID="0000-0002-1364-6789"))
             )
Description: Package for learning and evaluating (subgroup) policies via doubly robust loss functions. Policy learning methods include doubly robust blip/conditional average treatment effect learning and sequential policy tree learning. Methods for (subgroup) policy evaluation include doubly robust cross-fitting and online estimation/sequential validation. See Nordland and Holst (2026) <doi:10.18637/jss.v116.i04> for documentation and references.
License: Apache License (>= 2)
Encoding: UTF-8
Depends: R (>= 4.1), SuperLearner
Imports: data.table (>= 1.14.5), lava (>= 1.7.2.1), future.apply,
        progressr, methods, policytree (>= 1.2.0), survival, targeted
        (>= 0.6), DynTxRegime
Suggests: DTRlearn2, glmnet (>= 4.1-6), mets, mgcv, xgboost, knitr,
        ranger, rmarkdown, testthat (>= 3.0), ggplot2
RoxygenNote: 7.3.3
BugReports: https://github.com/AndreasNordland/polle/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-16 20:45:24 UTC; oano
Author: Andreas Nordland [aut, cre],
  Klaus Holst [aut] (ORCID: <https://orcid.org/0000-0002-1364-6789>)
Maintainer: Andreas Nordland <andreasnordland@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-17 22:50:02 UTC
Built: R 4.5.2; ; 2026-05-18 00:39:32 UTC; unix
