# Generated by roxygen2: do not edit by hand

S3method("+",policy_eval)
S3method(IC,policy_eval)
S3method(coef,policy_eval)
S3method(conditional,policy_eval)
S3method(estimate,policy_eval)
S3method(get_K,policy_data)
S3method(get_action_set,policy_data)
S3method(get_actions,policy_data)
S3method(get_event,history)
S3method(get_event,policy_data)
S3method(get_g_functions,policy_eval)
S3method(get_g_functions,policy_object)
S3method(get_history,policy_data)
S3method(get_history_names,policy_data)
S3method(get_id,history)
S3method(get_id,policy_data)
S3method(get_id_stage,history)
S3method(get_id_stage,policy_data)
S3method(get_n,policy_data)
S3method(get_policy,blip)
S3method(get_policy,drql)
S3method(get_policy,earl)
S3method(get_policy,owl)
S3method(get_policy,policy_eval)
S3method(get_policy,ptl)
S3method(get_policy,ql)
S3method(get_policy,rwl)
S3method(get_policy_actions,policy_eval)
S3method(get_policy_functions,blip)
S3method(get_policy_functions,drql)
S3method(get_policy_functions,policy_eval)
S3method(get_policy_functions,ptl)
S3method(get_policy_functions,ql)
S3method(get_policy_object,policy_eval)
S3method(get_q_functions,policy_eval)
S3method(get_q_functions,policy_object)
S3method(get_stage_action_sets,policy_data)
S3method(get_stage_data,policy_data)
S3method(get_utility,policy_data)
S3method(merge,policy_eval)
S3method(partial,policy_data)
S3method(plot,policy_data)
S3method(plot,policy_eval)
S3method(predict,QV_function)
S3method(predict,Q_function)
S3method(predict,SL.glm.formula)
S3method(predict,SL.nls)
S3method(predict,blip_function)
S3method(predict,c_cox)
S3method(predict,c_function)
S3method(predict,c_functions)
S3method(predict,c_no_censoring)
S3method(predict,g_empir)
S3method(predict,g_function)
S3method(predict,g_glm)
S3method(predict,g_glmnet)
S3method(predict,g_rf)
S3method(predict,g_sl)
S3method(predict,g_xgboost)
S3method(predict,m_function)
S3method(predict,nuisance_functions)
S3method(predict,q_degen)
S3method(predict,q_glm)
S3method(predict,q_glmnet)
S3method(predict,q_rf)
S3method(predict,q_sl)
S3method(predict,q_xgboost)
S3method(print,Q_function)
S3method(print,g_function)
S3method(print,nuisance_functions)
S3method(print,policy)
S3method(print,policy_data)
S3method(print,policy_eval)
S3method(print,policy_learn)
S3method(print,policy_object)
S3method(print,summary.policy_data)
S3method(subset_id,policy_data)
S3method(summary,policy_data)
S3method(summary,policy_eval)
S3method(vcov,policy_eval)
S3method(vcov,policy_eval_online)
export(All)
export(IC)
export(c_cox)
export(c_no_censoring)
export(conditional)
export(control_blip)
export(control_drql)
export(control_earl)
export(control_owl)
export(control_ptl)
export(control_rwl)
export(copy_policy_data)
export(estimate)
export(fit_c_functions)
export(fit_g_functions)
export(g_empir)
export(g_glm)
export(g_glmnet)
export(g_rf)
export(g_sl)
export(g_xgboost)
export(get_K)
export(get_action_set)
export(get_actions)
export(get_event)
export(get_g_functions)
export(get_history)
export(get_history_names)
export(get_id)
export(get_id_stage)
export(get_n)
export(get_policy)
export(get_policy_actions)
export(get_policy_functions)
export(get_policy_object)
export(get_q_functions)
export(get_stage_action_sets)
export(get_utility)
export(partial)
export(policy_data)
export(policy_def)
export(policy_eval)
export(policy_eval_online)
export(policy_learn)
export(q_glm)
export(q_glmnet)
export(q_rf)
export(q_sl)
export(q_xgboost)
export(sim_multi_stage)
export(sim_single_stage)
export(sim_single_stage_multi_actions)
export(sim_two_stage)
export(sim_two_stage_multi_actions)
export(subset_id)
import(SuperLearner)
import(data.table)
importFrom(SuperLearner,All)
importFrom(lava,IC)
importFrom(lava,estimate)
importFrom(methods,formalArgs)
importFrom(progressr,progressor)
importFrom(stats,.getXlevels)
importFrom(stats,addmargins)
importFrom(stats,aggregate)
importFrom(stats,as.formula)
importFrom(stats,binomial)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,delete.response)
importFrom(stats,gaussian)
importFrom(stats,glm)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,na.pass)
importFrom(stats,nls)
importFrom(stats,predict)
importFrom(stats,predict.glm)
importFrom(stats,reformulate)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(survival,Surv)
importFrom(survival,survfit)
importFrom(utils,capture.output)
importFrom(utils,tail)
