
<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![license](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/license/apache-2-0)
[![cran](https://www.r-pkg.org/badges/version-last-release/polle)](https://cranlogs.r-pkg.org/downloads/total/last-month/polle)
<!-- badges: end -->

# Policy Learning (`polle`) <img src="man/figures/logo.png" align="right" height="250" style="float:right; height:250px;"  alt="polle website" />

Package for evaluating user-specified finite stage policies and learning
optimal treatment policies via doubly robust loss functions. Policy
learning methods include doubly robust learning of the blip/conditional
average treatment effect and sequential policy tree learning. The
package also include methods for optimal subgroup analysis. See Nordland
and Holst (2026)
[doi:10.18637/jss.v116.i04](https://doi.org/10.18637/jss.v116.i04) for
documentation and references.
